Turninator = LibStub("AceAddon-3.0"):NewAddon("Turninator", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Turninator")

--==== set upvalues ====--
-- TODO: find out if QuestInfoFrame can be upvalued
-- GLOBALS: LibStub, CreateFrame, QuestInfoFrame, UIParent
local _G, Turninator, ceil, strfind, format, ipairs, pairs, print, select, strlen, strmatch, strtrim, tsort, tinsert, tonumber, type, unpack = _G, Turninator, ceil, strfind, format, ipairs, pairs, print, select, strlen, string.match, strtrim, table.sort, tinsert, tonumber, type, unpack
local AcceptQuest, CloseGossip, CloseQuest, CompleteQuest, DeclineQuest, GetActiveTitle, GetGossipActiveQuests, GetGossipAvailableQuests, GetGossipOptions, GetNumActiveQuests, GetNumAutoQuestPopUps, GetNumAvailableQuests, GetNumGossipAvailableQuests, GetNumGossipOptions, GetNumQuestChoices, GetQuestReward, IsQuestCompletable, QuestChooseRewardError, SelectActiveQuest, SelectAvailableQuest, SelectGossipActiveQuest, SelectGossipAvailableQuest, SelectGossipOption = AcceptQuest, CloseGossip, CloseQuest, CompleteQuest, DeclineQuest, GetActiveTitle, GetGossipActiveQuests, GetGossipAvailableQuests, GetGossipOptions, GetNumActiveQuests, GetNumAutoQuestPopUps, GetNumAvailableQuests, GetNumGossipAvailableQuests, GetNumGossipOptions, GetNumQuestChoices, GetQuestReward, IsQuestCompletable, QuestChooseRewardError, SelectActiveQuest, SelectAvailableQuest, SelectGossipActiveQuest, SelectGossipAvailableQuest, SelectGossipOption
local GetBindingByKey, GetBindingKey, GetItemInfo, IsUsableItem, SetBinding, UnitAffectingCombat, UnitCanAttack, UnitExists, UnitIsCorpse, UnitIsDead, UnitIsFriend, UnitName, UnitPlayerControlled = GetBindingByKey, GetBindingKey, GetItemInfo, IsUsableItem, SetBinding, UnitAffectingCombat, UnitCanAttack, UnitExists, UnitIsCorpse, UnitIsDead, UnitIsFriend, UnitName, UnitPlayerControlled


Turninator.DL = {
	[0] = "standard",
	"error",
	"test",
	"events",
	"config",
	"spam",
}
local DL = Turninator.DL
do	-- create fast lookups in table for reverse lookup and index number as string
	for i=0, #DL do
		DL[ DL[i] ] = i
		DL[tostring(i)] = i
	end
end

local dbgprint = { print   = function(...) return Turninator:Print(...) end,
                   squelch = function(...) return nil end }
local function SetDebugLevel(iDL)
	for i = 0,     iDL do dbgprint[ DL[i] ] = dbgprint.print;   end
	for i = iDL+1, #DL do dbgprint[ DL[i] ] = dbgprint.squelch; end
end

local tblCommands = { 
	DialogAction  = { strCommand = "CLICK TurninatorButton:LeftButton", strConfigMsg = "taking binding for dialog event" },
	PopUpQuest    = { strCommand = "CLICK WatchFrameAutoQuestPopUp1:LeftButton", strConfigMsg = "taking binding for quest popup" },
--[==[  -- TODO  LOOT
	LootConfirm   = { strCommand = "CLICK StaticPopup1Button1:LeftButton", strConfigMsg = "taking binding for loot confirm" },
--]==]  -- TODO  LOOT
	UseItem       = { strCommand = nil, strConfigMsg = "taking binding for item use" },
	Interact      = { strCommand = "INTERACTTARGET", strConfigMsg = "taking binding for interaction" },
	Released      = { strCommand = nil, strConfigMsg = "releasing binding" },
}

local CEventHook = { module = Turninator, isAttached = false }
CEventHook.__index = CEventHook
function CEventHook:Set(fAttach)
	if self.isAttached ~= fAttach then
		if fAttach then
			self.Attach(self.module, self.name, self.handler)
		else
			self.Detach(self.module, self.name)
		end
		self.isAttached = fAttach
	end
end
local function CreateEventHook(type, name, handler)
	if type == "Event" then
		return setmetatable( { name = name, handler = handler or name, Attach = CEventHook.module.RegisterEvent, Detach = CEventHook.module.UnregisterEvent }, CEventHook )
	elseif type == "Hook" then
		return setmetatable( { name = name, handler = handler or name, Attach = CEventHook.module.SecureHook, Detach = CEventHook.module.Unhook }, CEventHook )
	end
end

local tblMainEventHooks = {
	CreateEventHook("Event", "GOSSIP_SHOW",           "DialogEventHandler"),
	CreateEventHook("Event", "GOSSIP_CLOSED",         "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_GREETING",        "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_DETAIL",          "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_PROGRESS",        "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_COMPLETE",        "DialogEventHandler"),
	CreateEventHook("Event", "QUEST_FINISHED",        "DialogEventHandler"),
	CreateEventHook("Event", "PLAYER_TARGET_CHANGED", "SelectBinding"),
--	CreateEventHook("Event", "UPDATE_BINDINGS",       "UpdateBindingsHandler"),
	CreateEventHook("Hook",  "AddAutoQuestPopUp",     "AddAutoQuestPopUpHook"),
	CreateEventHook("Hook",  "RemoveAutoQuestPopUp",  "RemoveAutoQuestPopUpHook"),

--[==[  -- TODO  LOOT
	CreateEventHook("Event", "LOOT_BIND_CONFIRM",     "LootConfirm"),
--	CreateEventHook("Event", "LOOT_OPENED",           "LootOpened"),
	CreateEventHook("Event", "LOOT_READY",            "LootOpened"),
	CreateEventHook("Event", "LOOT_CLOSED",           "LootClosed"),
	CreateEventHook("Event", "LOOT_SLOT_CLEARED",     "LootSlotCleared"),
	CreateEventHook("Event", "LOOT_SLOT_CHANGED",     "LootSlotChanged"),
--]==]  -- TODO  LOOT
}

--[==[  -- TODO  LOOT
local function HandleLootClick(tblLoot)
	local strFrame = StaticPopup_Visible("LOOT_BIND")
	if strFrame ~= nil then
		_G[strFrame.."Button1"]:Click("LeftButton", false)
	else
		local slot = tblLoot and next(tblLoot)
		if slot then LootSlot(slot) end
	end
end

local tblLoot = {}

function Turninator:LootConfirm(event, slot, ...)
	if GetNumGroupMembers() == 0 then
		if type(slot) ~= "number" or GetNumLootItems() < slot or not LootSlotHasItem(slot) then  -- sanity check, probably not needed
			dbgprint.error("sanity check failed", event, slot, ...)
		else
			dbgprint.test(event, slot, ...)
			tblLoot[slot] = true
			self.tblDialogAction = { HandleLootClick, tblLoot } 
			self:SelectBinding(event);
		end	
	end
end

function Turninator:LootSlotCleared(event, slot, ...)
	dbgprint.test(event, slot, ...)
	tblLoot[slot] = nil
end

function Turninator:LootSlotChanged(event, ...)
	dbgprint.test(event, ...)
end

function Turninator:LootOpened(event, ...)
	dbgprint.test(event, ...)
end

function Turninator:LootClosed(event, ...)
	dbgprint.test(event, ...)
	tblLoot = {}
	self.tblDialogAction = {} 
	self:SelectBinding(event)
end
--]==]  -- TODO  LOOT


local tblZoneEventHooks = {
	CreateEventHook("Event", "ZONE_CHANGED_NEW_AREA", "UpdateHooks"),
	CreateEventHook("Event", "PLAYER_ENTERING_WORLD", "UpdateHooks"),
}


-- Called when the addon is loaded
function Turninator:OnInitialize()
	Turninator:LoadSavedSettings()

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Turninator", Turninator.GetUIOptions, { "Turninator", "tr" } )
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Turninator", "Turninator")

	for k,v in pairs(self.DL) do DL[v] = k; DL[k] = v end
	SetDebugLevel(self.db.global.iDebugLevel)

	-- if there's no shared command, then have the binding release to Interact with Target
	tblCommands.Released.strCommand = self.db.profile.strSharedCmd or tblCommands.Interact.strCommand
	self.strCurBind = "Released"
	self.tblDialogAction = {}
end

function Turninator:UpdateHooks(fForce)
	fHookMains = self.db.profile.fEnabled and not (self.db.profile.fDisableInInstances and IsInInstance())
	dbgprint.spam(fHookMains and "enabling addon" or "disabling addon")
	for _, hook in pairs(tblMainEventHooks) do
		hook:Set(fHookMains)
	end

	local fHookZones = self.db.profile.fEnabled and self.db.profile.fDisableInInstances
	for _, hook in pairs(tblZoneEventHooks) do
		hook:Set(fHookZones)
	end

	self:SelectBinding("UpdateHooks")
end

-- Called when the addon is enabled
function Turninator:OnEnable()
	self:CreateButton()
	self:UpdateHooks()
end

-- Called when the addon is Disabled
function Turninator:OnDisable()
	if (self.db.profile.strSharedKey) then
		SetBinding(self.db.profile.strSharedKey, self.db.profile.strSharedCmd);
	end

	for _, hook in pairs(tblMainEventHooks) do
		hook:Set(false)
	end
	for _, hook in pairs(tblZoneEventHooks) do
		hook:Set(false)
	end
end

function Turninator:SetEnabledState(info, value)
	value = value or (info and info[#info])
	if (value == L["On"]) then
		self.db.profile.fEnabled = true
		self:UpdateHooks()
	elseif (value == L["Off"]) then
		self.db.profile.fEnabled = false
		self.tblDialogAction = {}
		self:UpdateHooks()
	else
		dbgprint.standard("ERROR: no valid state passed to SetEnabledState");
	end
	
	self:Config(info)
end

function Turninator:CreateButton()
	if (not self.btn) then
		self.btn = CreateFrame("Button", "TurninatorButton", UIParent, "SecureActionButtonTemplate");
		self.btn:SetAttribute("type", "macro")
		self.btn:SetAttribute("macrotext", "/run Turninator:DialogAction()")
	end
end


function Turninator:EventHandler(...)
	dbgprint.events(...)
end

function Turninator:UpdateBindingsHandler(...)
	local strBind = self.db.profile.strSharedKey and GetBindingByKey(self.db.profile.strSharedKey) or ""
	dbgprint.events(..., strBind)
end

function Turninator:DialogEventHandler(event, ...)
	dbgprint.events(event, ...)
	self.tblDialogAction = self:GetDialogAction(event, ...)
	self:SelectBinding(event);
end

function Turninator:AddAutoQuestPopUpHook(...) -- questId, type)
	dbgprint.events("AddAutoQuestPopUpHook", ...)
	self:SelectBinding("AddAutoQuestPopUpHook")
end

function Turninator:RemoveAutoQuestPopUpHook(...) -- questId
	dbgprint.events("RemoveAutoQuestPopUpHook", ...)
	self:SelectBinding("RemoveAutoQuestPopUpHook")
end


function Turninator:Config(info)
	if (info and info.uiType == "cmd") then
		LibStub("AceConfigRegistry-3.0"):NotifyChange("Turninator")
		if (info[#info] == L["Config"]) then 
			_G.InterfaceOptionsFrame_OpenToCategory(Turninator.optionsFrame)
		else
			dbgprint.error("debug reporting level set to", self.db.global.iDebugLevel, DL[self.db.global.iDebugLevel])
			if (self.db.profile.fEnabled and self.db.profile.fDisableInInstances and IsInInstance()) then
				dbgprint.standard(L["addon is"], L["temporarily disabled for an instance"])
			else
				dbgprint.standard(L["addon is"], self.db.profile.fEnabled and L["enabled"] or L["disabled"])
			end
			dbgprint.standard(L["shared cmd is"], self.db.profile.strSharedCmd or L["not set"])
			dbgprint.standard(L["shared key is"], self.db.profile.strSharedKey or L["not set"])
			if (self.db.profile.strSharedKey) then 
				dbgprint.spam("shared key currently bound to", GetBindingByKey(self.db.profile.strSharedKey))
			end
		end
	end
end


function Turninator:Debug(info, value)
	local strValue, index = self:GetArgs(value, 1);

	if (strValue ~= nil) then
		if (index ~= 1e9) then
			dbgprint.standard(L["ERROR: too many parameters passed"]);
		end

		local iValue = DL[strValue]
		if (iValue == nil) then
			dbgprint.standard("ERROR: value must be number between 0 and", #DL);
			return;
		else
			self.db.global.iDebugLevel = iValue
			SetDebugLevel(iValue)
		end
	end

	self:Config(info)
end

function Turninator:SetIconCallback(idItem)
	local strItemName, _, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(idItem)
	if not strItemName then
		dbgprint.standard("ERROR: invalid item id: ", idItem);
	else
		dbgprint.standard("second attempt successful for ", idItem);
		self.db.global.tblItemIcon[strItemName] = strItemIcon
	end
end

function Turninator:SetIconById(info, value)
	local strValue, index = self:GetArgs(value, 1, 1)
	while (strValue~= nil) do
		local idItem = tonumber(strValue)
		if not idItem then
			dbgprint.standard("ERROR: non-numeric argument passed: ", strValue);
		else
			local strItemName, _, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(idItem)
			if not strItemName then
				self:ScheduleTimer("SetIconCallback", 1, idItem);
				-- dbgprint.standard("ERROR: invalid item id: ", idItem);
			else
				self.db.global.tblItemIcon[strItemName] = strItemIcon
			end
		end
		strValue, index = self:GetArgs(value, 1, index)
	end
end

function Turninator:Reset(info)
	if (self.db.profile.strSharedKey) then
		SetBinding(self.db.profile.strSharedKey, self.db.profile.strSharedCmd);
	end

	self.db.profile.strSharedKey = nil
	self.db.profile.strSharedCmd = nil

	tblCommands.Released.strCommand = tblCommands.Interact.strCommand
	self.strCurBind = "Released"
	self.tblDialogAction = {}

	self:Config(info)
end

function Turninator:ShareKey(info, value)
	local strKey, index = self:GetArgs(value, 1);

	if (strKey == nil) then
		if (info and info.uiType == "cmd") then
			dbgprint.standard(L["ERROR: no key specified"]);
		else
			Turninator:Reset(info)
		end
		return
	elseif (index ~= 1e9) then
		dbgprint.standard(L["ERROR: too many parameters passed"]);
		return
	end

	local strAction = GetBindingByKey(strKey)
	if (not SetBinding(strKey, strAction)) then
		dbgprint.standard(L["ERROR: invalid key"])
		return;
	end

	self:Reset();  -- release old binding before setting up new one

	if (not strAction or strAction == "") then
		dbgprint.config("sharing previously unbound key");
	else
		dbgprint.config("sharing key", strKey, "with action", strAction);
		self.db.profile.strSharedCmd = strAction
		tblCommands.Released.strCommand = strAction
	end
	self.db.profile.strSharedKey = strKey

	self.strCurBind = "Released"
	self:SelectBinding("ShareKey")
	self:Config(info)
end

function Turninator:ShareCmd(info, value)
	local strAction, index = self:GetArgs(value, 1);

	if (strAction == nil) then
		dbgprint.standard(L["ERROR: no action specified"]);
		return;
	elseif (index ~= 1e9) then
		dbgprint.standard(L["ERROR: too many parameters passed"]);
		return
	end

	local strKey = GetBindingKey(strAction)
	if (not strKey or strKey == "") then
		dbgprint.standard(L["ERROR: action does not exist or is unbound"]);
		return;
	end

	self:Reset();  -- release old binding before setting up new one

	self.db.profile.strSharedCmd = strAction
	self.db.profile.strSharedKey = strKey

	tblCommands.Released.strCommand = strAction
	self.strCurBind = "Released"
	self:SelectBinding("ShareCmd")
	self:Config(info)
end

function Turninator:UseItem(info, strItem)
	local strName = UnitName("target")
	if (not strName or UnitPlayerControlled("target")) then
		dbgprint.standard(L["ERROR: please target an NPC when using this command"])
		return
	elseif (strItem == nil) then
		dbgprint.standard(L["ERROR: no item specified"]);
		return
	end

	local strItemName, _, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(strItem)
	if (strItemName == nil) then
		dbgprint.standard(L["ERROR: item not found in inventory"]);
		return
	end

	-- GET_ITEM_INFO_RECEIVED
	-- ITEM_PUSH

	dbgprint.standard("adding item", strItemName, "for use on", strName)
	self.db.global.tblUseItem[strName] = strItemName
	self.db.global.tblItemIcon[strItemName] = strItemIcon
	self:TakeBinding("Released", "UseItem")
	self:SelectBinding("UseItem")
end

function Turninator:SetNPC(info)
	local strAction = info and info[#info]
	local strName = UnitName("target")

	if (not strName or UnitPlayerControlled("target")) then
		dbgprint.standard(L["ERROR: please target an NPC when using this command"])
	elseif (strAction == L["Interact"]) then
		dbgprint.standard("adding", strName, "to Interact list")
		self.db.global.tblInteract[strName] = true
		self:SelectBinding("Interact")
	elseif (strAction == L["Ignore"]) then
		dbgprint.standard("adding", strName, "to Ignore list")
		self.db.global.tblInteract[strName] = false
		self:SelectBinding("Ignore")
	elseif (strAction == L["Forget"]) then
		dbgprint.standard("forgetting", strName)
		self.db.global.tblInteract[strName] = nil
		self.db.global.tblUseItem[strName] = nil
		self:SelectBinding("Forget")
	end

	LibStub("AceConfigRegistry-3.0"):NotifyChange("Turninator");
end

function Turninator:ListNPCs(info)
	local ignore, interact, items = {}, {}, {}
	local cmac = 0

	for strNPCName,fInteract in pairs(self.db.global.tblInteract) do
		tinsert( (fInteract and interact or ignore), strNPCName)
	end

	if (#interact > 0) then
		dbgprint.standard(L["forcing interaction with:"])
		tsort(interact)
		for _,v in ipairs(interact) do
			print(format("  %s", v))
		end
	end

	if (#ignore > 0) then
		dbgprint.standard(L["ignoring:"])
		tsort(ignore)
		for _,v in ipairs(ignore) do
			print(format("  %s", v))
		end
	end


	for strNPCName,strItemName in pairs(self.db.global.tblUseItem) do
		tinsert( items, { name = strNPCName, item = strItemName } )
		local clen = strlen(strNPCName)
		if (cmac < clen) then cmac = clen end
	end

	if (#items) then 
		dbgprint.standard(L["npc item bindings:"])
		tsort(items, function(a,b) return (a.name < b.name) end)
		for _,v in ipairs(items) do
			print(format( format("  %%-%is%%s", cmac+3), v.name, v.item ))
		end
	end
end


function Turninator:SelectBinding(event)
	if (self.db.profile.strSharedKey == nil) then 
		if (self.db.profile.fEnabled) then dbgprint.error("ERROR: no key is set; cannot respond to", event); end
	elseif (self.db.profile.fEnabled and #self.tblDialogAction ~= 0) then
		self:TakeBinding("DialogAction", event)
	elseif (self.db.profile.fEnabled and GetNumAutoQuestPopUps() ~= 0) then
		self:TakeBinding("PopUpQuest", event)
	elseif (self.db.profile.fEnabled and self:CanUseItem()) then
		self:TakeBinding("UseItem", event)
	elseif (self.db.profile.fEnabled and self:CanInteract()) then
		self:TakeBinding("Interact", event)
	else
		self:TakeBinding("Released", event)
	end
end

function Turninator:TakeBinding(strBind, event)
	if (self.strCurBind ~= strBind) then
		if (UnitAffectingCombat("player")) then
			dbgprint.error("currently in combat; binding cannot be changed")
		else
			local bind = tblCommands[strBind]
			if (SetBinding(self.db.profile.strSharedKey, bind.strCommand)) then
				dbgprint.config(bind.strConfigMsg, event);
				self.strCurBind = strBind
			end
		end
	end
end


Turninator.tblProblemTurnins = {
	[87706] = { 
			[L["Aogexon's Fang"]] = { itemId = 118654, count = 1 },
			[L["Bergruu's Horn"]] = { itemId = 118655, count = 1 },
		   [L["Dekorhan's Tusk"]] = { itemId = 118656, count = 1 },
		   [L["Direhoof's Hide"]] = { itemId = 118657, count = 1 },
			[L["Gagrog's Skull"]] = { itemId = 118658, count = 1 },
			 [L["Mu'gra's Head"]] = { itemId = 118659, count = 1 },
		[L["Thek'talon's Talon"]] = { itemId = 118660, count = 1 },
		   [L["Vileclaw's Claw"]] = { itemId = 120172, count = 1 },
		[L["Xelganak's Stinger"]] = { itemId = 118661, count = 1 },
	},
	[87393] = { 
				  [L["A Rare Find"]] = { itemId = 118100, count = 1  },  -- highmaul relic
		[L["Fragments of the Past"]] = { itemId = 118099, count = 20 },  -- gorian artifact fragment
	},
};

function Turninator:GetDialogAction(event, ...)

	if (event == "GOSSIP_SHOW") then
		local function GetActive(...)
			local tbl = {}
			for iOffset = 1, select("#", ...), 5 do
				local iActive = ceil(iOffset/5)
				local strQuest = select(iOffset, ...)
				local fComplete = select(iOffset+3, ...)
				tbl[iActive] = { name = strQuest, complete = fComplete }
			end
			return tbl;
		end

		local function GetAvailable(...)
			local function CompareAvailableQuests(a, b)
				if (not (a.precedence == b.precedence)) then
					return a.precedence > b.precedence
				elseif (not (a.level == b.level)) then
					return a.level > b.level
				else
					return a.index < b.index
				end
			end

			local tbl = {}
			for iOffset = 1, select("#", ...), 6 do
				local index = ceil(iOffset/6)
				local strQuest, level, isTrivial, frequency, isRepeatable, isLegendary = select(iOffset, ...)
				local precedence = (isRepeatable and 0) or (isTrivial and 1) or (frequency ~= -1 and 2) or 3
				tinsert(tbl, { index = index, name = strQuest, level = level, precedence = precedence } )
			end
			tsort(tbl, CompareAvailableQuests)
			return tbl;
		end

		local function GetGossip(...)
			local tbl = {}
			for iOffset = 1, select("#", ...), 2 do
				local index = ceil(iOffset/2)
				local strGossip, strType = select(iOffset, ...)
				if (strType == "gossip") then
					tinsert(tbl, { index = index, text = strGossip } )
				end
			end
			return tbl;
		end

		-- check for quests we can turn in
		local tblActive = GetActive(GetGossipActiveQuests())
		for i = 1, #tblActive do
			if (tblActive[i].complete) then
				-- found a quest we can turn in, that's what button will do
				dbgprint.spam("will turn in quest", i, tblActive[i].name);
				return { SelectGossipActiveQuest, i }
			end
		end

		-- no quests we can turn in, so how about pick up an available quest
		local cAvailable = GetNumGossipAvailableQuests()
		dbgprint.spam(cAvailable, "available quests for this npc")
		if (cAvailable > 0) then
			local tblAvailable = GetAvailable(GetGossipAvailableQuests())

			npc = tonumber(strmatch(UnitGUID("target"), "Creature%-0%-%w+%-%w+%-%w+%-(%w+)%-%w+"))
			local turnins = self.tblProblemTurnins[npc]
			if not turnins then
				return { SelectGossipAvailableQuest, tblAvailable[1].index }
			else
				for _, quest in ipairs(tblAvailable) do
					local turnin = turnins[quest.name]
					if turnin and turnin.count <= GetItemCount(turnin.itemId) then
						return { SelectGossipAvailableQuest, quest.index }
					end		
				end
			end
		end

		-- if the npc has chatty gossip options try the first one
		if (GetNumGossipOptions() > 0) then
			local tblGossip = GetGossip(GetGossipOptions())
			dbgprint.spam(#tblGossip, "gossip text options for this npc")
			-- PrintTable(tblGossip)
			for _, tblChoice in pairs(tblGossip) do
				if (strfind(tblChoice.text, L["Darkmoon Game Token"])) then
					return { SelectGossipOption, tblChoice.index }
				elseif (strfind(tblChoice.text, L["Take me to the faire staging area"])) then
					return { SelectGossipOption, tblChoice.index, nil, true}
				elseif (strfind(tblChoice.text, L["Teleport me to the cannon"])) then
					return { SelectGossipOption, tblChoice.index, nil, true}
				end
			end
			if (tblGossip[1] ~= nil) then 
				return { SelectGossipOption, tblGossip[1].index }
			end
		end

		-- nothing we can do with this quest giver apparently ... exit seems iffy ... revisit
		return { CloseGossip }

	elseif (event == "QUEST_GREETING") then
		local function GetActive()
			local tbl = {}
			for iActive = 1, GetNumActiveQuests() do
				local strQuest, fComplete = GetActiveTitle(iActive)
				tbl[iActive] = { name = strQuest, complete = fComplete }
			end
			return tbl;
		end

		local tblActive = GetActive()
		for i = 1, #tblActive do
			if (tblActive[i].complete) then
				-- found a quest we can turn in, that's what button will do
				dbgprint.spam("will turn in quest", i, tblActive[i].name);
				return { SelectActiveQuest, i }
			end
		end

		-- no quests we can turn in, so how about pick up an available quest
		local cAvailable = GetNumAvailableQuests()
		dbgprint.spam(cAvailable, "available quests for this npc")
		if (cAvailable > 0) then
			return { SelectAvailableQuest, 1 }
		end

		-- nothing we can do with this quest giver apparently ... exit seems iffy ... revisit
		return { CloseQuest }

	elseif (event == "QUEST_DETAIL") then
		return { AcceptQuest }

	elseif (event == "QUEST_PROGRESS") then
		if (IsQuestCompletable()) then
			return { CompleteQuest }
		else
			return { DeclineQuest }
		end
		-- todo - can't complete, back up?

	elseif (event == "QUEST_COMPLETE") then
		local function CheckRewardSelection()
			if (QuestInfoFrame.itemChoice ~= 0) then
				GetQuestReward(QuestInfoFrame.itemChoice)
			else
				QuestChooseRewardError()
			end 
		end

		local numRewardChoices = GetNumQuestChoices()
		if (numRewardChoices == 0) then
			return { GetQuestReward }
		elseif (numRewardChoices == 1) then
			return { GetQuestReward, 1 }
		else
			return { CheckRewardSelection }
		end

	elseif (event == "GOSSIP_CLOSED") then
		return {}

	elseif (event == "QUEST_FINISHED") then
		return {}
	end

	return {}
end

function Turninator:DialogAction()
	if (not self.tblDialogAction or type(self.tblDialogAction) ~= "table") then
		dbgprint.error("ERROR - tblDialogAction is of type", type(self.tblDialogAction));
	elseif (not self.tblDialogAction[1] or type(self.tblDialogAction[1]) ~= "function") then
		dbgprint.error("ERROR - tblDialogAction does not contain function");
	else
		self.tblDialogAction[1](unpack(self.tblDialogAction, 2));
	end
end

local tblItemIDFixups = {
	[74258] = 74771,  -- Staff of Pei-Zhi
}

function Turninator:CanUseItem()
	local name = UnitName("target")
	if (not name or UnitPlayerControlled("target")) then
		return false
	end

	local item = self.db.global.tblUseItem[name]
	if (not item) then
		return false
	end
	
	 if not IsUsableItem(item) then
		local _, strItemLink, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(item)
		local idItem = strItemLink and strmatch(strItemLink, "item:(%d+)")
		local idFixed = tblItemIDFixups[tonumber(idItem)]
		if not idFixed or not IsUsableItem(idFixed) then
			return false
		end
	end

	-- if we're missing the icon path (association created by v1 of the addon) grab it now
	if (not self.db.global.tblItemIcon[item]) then
		local strItemName, _, _, _, _, _, _, _, _, strItemIcon = GetItemInfo(item)
		if (item == strItemName) then
			self.db.global.tblItemIcon[strItemName] = strItemIcon
		end
	end

	-- set the binding up in tblCommands
	tblCommands.UseItem.strCommand = format("ITEM %s", item)
	-- print(tblCommands.UseItem.strCommand)
	return true
end

function Turninator:CanInteract()
	local name = UnitName("target")
	if name == "Human Orphan" or name == "Draenei Orphan" then return true end

	if (not name or UnitPlayerControlled("target")) then
		return false
	end

	local interact = self.db.global.tblInteract[name]
	if (interact ~= nil) then
		return interact
	end

	return (UnitIsFriend("player", "target") and not UnitIsCorpse("target") 
		and not UnitIsDead("target") and not UnitCanAttack("player", "target"))
end

